"""
A feedforward neural network. Weights are updated using stochastic gradient descent, and gradients are calculated using backpropagation.

The code is based on [1] and [2].

[1] https://github.com/mnielsen/neural-networks-and-deep-learning
[2] http://deeplearning.net/tutorial/mlp.html
"""

import pickle as pickle
import numpy as np

rng = np.random.RandomState()

# The sigmoid function
def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))

# The gradient (derivative) of the sigmoid function
def sigmoid_prime(x):
    s = sigmoid(x)
    return s * (1.0 - s)

# The derivative of the error function (here: mean squared error) for comparing
# an observed value "y" to a target value "t". (This is the partial derivative
# of the function 1/2 (t - y)^2 with respect to y.)
def error_prime(y, t):
    return y - t

# A simple feedforward network. Every network has one or more layers. The first
# layer is the input layer; the neurons on this layer are just buffers for
# input values. Each neuron on a subsequent layer receives weighted inputs from
# all neurons on the preceding layer. Each such neuron also has a bias.
class Network(object):
    
    def __init__(self, sizes):
        """
        Initialize the network. The list "sizes" specifies the number of
        neurons per layer.
        """
        self.sizes = sizes
        # Initialize biases and weights. The field self.b is a list of column
        # vectors, one vector for each layer but the input layer. The field
        # self.w is a list of m x n matrices, one for each layer but the input
        # layer, where m is the number of nodes on the current layer and n is
        # the number of nodes on the previous layer. The entry w_cp specifies
        # the weight of connection to node c on the current layer from node p
        # on the previous layer. By default, biases and weights are initialized
        # using random samples from the standard normal distribution.
        self.b = [rng.randn(c, 1)
            for c in self.sizes[1:]]
        self.w = [rng.randn(c, p)
            for c, p in zip(self.sizes[1:], self.sizes[:-1])]
        # The (vectorized) activation function; here, the sigmoid function.
        self.f = np.vectorize(sigmoid)
        # The (vectorized) derivative of the activation function.
        self.f_prime = np.vectorize(sigmoid_prime)
    
    def init_weights_sigmoid(self):
        """
        Initializes the weights using random samples from the uniform
        distribution over the symmetric interval suggested by Glorot and
        Bengio (2010), as described here:
        http://deeplearning.net/tutorial/mlp.html
        """
        self.b = [rng.uniform(
            low=-4 * np.sqrt(6.0 / (c + 1)),
            high=4 * np.sqrt(6.0 / (c + 1)),
            size=(c, 1)
        ) for c in self.sizes[1:]]
        self.w = [rng.uniform(
            low=-4 * np.sqrt(6.0 / (c + p)),
            high=4 * np.sqrt(6.0 / (c + p)),
            size=(c, p)
        ) for c, p in zip(self.sizes[1:], self.sizes[:-1])]
    
    def feedforward(self, x):
        """
        Returns the output of the network for input "x".
        """
        for b, w in zip(self.b, self.w):
            x = self.f(np.dot(w, x) + b)
        return x
    
    def update(self, batch, eta):
        """
        Updates the network weights using the gradient descent update rule.
        This uses backpropagation to compute the gradients for the biases and
        the weights, based on the data in "batch". The parameter "eta"
        specifies the learning rate.
        """
        nabla_b = [np.zeros(b.shape) for b in self.b]
        nabla_w = [np.zeros(w.shape) for w in self.w]
        for x, y in batch:
            delta_nabla_b, delta_nabla_w = self.backpropagate(x, y)
            nabla_b = [nb + dnb for nb, dnb in zip(nabla_b, delta_nabla_b)]
            nabla_w = [nw + dnw for nw, dnw in zip(nabla_w, delta_nabla_w)]
        self.b = [b - (eta / len(batch)) * nb for b, nb in zip(self.b, nabla_b)]
        self.w = [w - (eta / len(batch)) * nw for w, nw in zip(self.w, nabla_w)]
    
    def backpropagate(self, x, t):
        """
        Computes the gradients for the biases and the weights given the current
        error. The error is computed as the mean squared error between the
        output of the network on input "x" and the target output value "t".
        (Note that by "mean squared error" we mean the fraction of the mean
        squared error contributed by one training sample.)
        """
        # Forward pass. Compute weighted inputs (zs) and activations (ys) for
        # each layer of the network.
        zs = []
        ys = [x]
        for b, w in zip(self.b, self.w):
            z = np.dot(w, x) + b
            x = self.f(z)
            zs.append(z)
            ys.append(x)
        # Backward pass. Compute the error (delta) and the rates of change of
        # the cost with respect to any bias (nabla_b) and weight (nabla_w) in
        # the network.
        delta = error_prime(ys[-1], t) * self.f_prime(zs[-1])
        nabla_b = [delta]
        nabla_w = [np.dot(delta, ys[-2].transpose())]
        for l in range(2, len(self.sizes)):
            delta = np.dot(self.w[-l+1].transpose(), delta) * self.f_prime(zs[-l])
            nabla_b.append(delta)
            nabla_w.append(np.dot(delta, ys[-l-1].transpose()))
        return (nabla_b[::-1], nabla_w[::-1])
    
    def save(self, file):
        """
        Saves this network to "file".
        """
        with open(file, 'wb') as f:
            pickle.dump(self.sizes, f, pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.b, f, pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.w, f, pickle.HIGHEST_PROTOCOL)
    
    @staticmethod
    def load(file):
        """
        Loads a network from "file".
        """
        net = Network([])
        with open(file, 'rb') as f:
            net.sizes = pickle.load(f, encoding='latin1')
            net.b = pickle.load(f, encoding='latin1')
            net.w = pickle.load(f, encoding='latin1')
        return net

    def set_seed(seed=1234):
        # The random number generator used in this code. In order to make results
        # replicable, the generator is initialized with a specific seed.
        rng = np.random.RandomState(seed)

def train(network, data, n_epochs, batch_size, eta, evaluate=None):
    """
    Trains a network using minibatches. The "data" is a list of pairs "(x, y)",
    each of which represents a training input and the desired output. The
    parameter "n_epochs" specifies the number of epochs for which the network
    is trained; the parameter "batch_size" specifies the number of items per
    minibatch; the parameter "eta" is the learning rate. Finally, the optional
    parameter "evaluate" is a function that computes the error rate of the
    network based on some testing data. If provided, this function will be
    run after each epoch and its result will be printed.
    """
    n = len(data)
    for e in range(n_epochs):
        rng.shuffle(data)
        for k in range(0, n, batch_size):
            mini_batch = data[k:k+batch_size]
            network.update(mini_batch, eta)
            #print "\rEpoch %02d, %05d instances" % (e, k + batch_size),

        if evaluate:
            error_rate = evaluate(network)
            print("Epoch %02d, eta = %.2f, error rate = %.2f" % (e, eta, error_rate * 100))
